#
# Copyright (c) 2017 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..artefact import artefact, source, notfile
from ..tools.installer import installer, prefix, stage  # noqa F401
from ..rule import rule, alias
from .. import platform
from os.path import normpath, join, splitdrive

default_prefix = prefix('/usr/local')
if platform.os == 'Windows':
    default_prefix = prefix(r'C:/Program Files')


class _installed(artefact):
    """An installed artefact has a filename outside the build directory."""

    def __init__(self, a, subdir, features):
        """Create an installed artefact."""
        a = source.instantiate(a)
        self.a = a
        self.subdir = subdir or ''
        artefact.__init__(self, a.name, type=a.type, module=a.module, features=features)
        # set a default installation prefix
        if 'prefix' not in self.features:
            self.features += default_prefix
        self._define_rule()

    def __call__(self, features):
        clone = artefact.__call__(self, features)
        clone._define_rule()
        return clone

    @property
    def relpath(self):
        # combine stage with prefix to form the actual path
        stage = str(self.features.get('stage', ''))  # noqa F811
        prefix = str(self.features.get('prefix', ''))  # noqa F811
        if stage:
            if stage[-1] != '/':
                stage += '/'
            prefix = splitdrive(prefix)[1]
        return join(stage + prefix, self.subdir)

    @property
    def _filename(self):
        # use the name of the upstream artefact, combined with
        # our relpath
        return normpath(join(self.relpath, self.a.name))

    def _define_rule(self):
        rule(installer.install, self, self.a)


def installed(artefact, subdir=None, features=()):
    """Make an installed copy of the given artefact."""

    from . import library
    # TODO: Find an extensible mechanism to make this polymorphic
    if isinstance(artefact, library.library):
        return library.installed(artefact, subdir, features)
    else:
        return _installed(artefact, subdir, features)


class manifest(artefact):

    def create(t, sources):
        stage = str(t[0].features.stage) if 'stage' in t[0].features else ''  # noqa F811
        files = [s._filename[len(stage):] for s in sources
                 if s._filename.startswith(stage)]
        with open(t[0]._filename, 'w') as manifest:
            for f in files:
                manifest.write(f + '\n')
    create.__name__ = 'manifest.create'
    create = staticmethod(create)

    def __init__(self, name, installed, features=()):
        # HACK: this works around a Windows filesystem limitation
        # TODO: Find a real solution for this, including establishing
        #       and documenting artefact naming requirements, etc.
        name = name.replace(':', '_')
        artefact.__init__(self, name, features=features)
        rule(manifest.create, self, installed)


class installation(artefact):

    def __init__(self, name, installed, features=()):
        artefact.__init__(self, name, attrs=notfile, features=features)
        self.installed = installed
        self.manifest = manifest(self.name + '.manifest', installed, features=features)
        alias(self, self.manifest)

    def __call__(self, features):
        clone = artefact.__call__(self, features)
        clone.installed = [i(features) for i in self.installed]
        clone.manifest = manifest(self.name + '.manifest', clone.installed, features=features)
        alias(clone, clone.manifest)
        return clone



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..feature import set
from ..artefacts.library import library
from ..tools.compiler import link, ldflags
from ..tools.python import python
from os.path import join, normpath


class extension(library):

    def __init__(self, *args, **kwds):
        library.__init__(self, *args, **kwds)
        p = python.instance(self.features)
        self.features |= set(p.include, p.linkpath, link('shared'))
        # on windows we need to link with libpython
        self.features |= p.libs(condition=(set.cc.name=='msvc')|(set.cxx.name=='msvc'))
        # on darwin we need to add `-undefined dynamic_lookup` to prevent undefined symbols errors
        self.features |= ldflags('-undefined dynamic_lookup', condition=(set.target.os.matches('darwin.*')))
        self._suffix = p.ext_suffix

    @property
    def _filename(self):
        libname = self.name + self._suffix
        return normpath(join(self.module.builddir, self.relpath, libname))



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..artefact import artefact, notfile, nocare, always
from ..action import action
from ..feature import map, join
from ..rule import rule, depend
from ..tools import compiler
from .. import platform
from .. import output
from ..utils import aslist
from . import pass_, fail, xpass, xfail
import logging

logger = logging.getLogger('actions')


class test(artefact):
    """A test is an artefact that is typically performed as part of a test suite, with
    special support to collect and report the test outcome."""

    class run(action):

        runpath = map(compiler.runpath, join)
        if platform.os == 'Windows':
            command = """set PATH=$(runpath);%PATH%
$(>)"""
        else:
            command = 'LD_LIBRARY_PATH=$(runpath) $(>)'

    def __init__(self, name, sources, run=False, dependencies=[], features=(),
                 condition=None, expected=pass_, module=None):
        """Create a test.

        Arguments:
          * name: the test's name
          * sources: artefact(s) to be tested
          * run: Specify what to do. Possible values:

             - `False` (default): do nothing (but report whether source was updated successfully)
             - `True`: Run sources
             - an action: Perform action

          * depends: any prerequisite artefacts that have to be completed
          * features: additional features needed to perform this test
          * condition: either a boolean or a feature condition to indicate if this test is to be performed or skipped.
          * expected: the expected outcome"""

        artefact.__init__(self, name, attrs=notfile|nocare|always, features=features, condition=condition, module=module)
        sources = aslist(sources)
        self.xoutcome = expected
        self.outcome = None
        self.command = ''
        self.output = ('', '')
        if run is True:
            rule(test.run(), self, sources, dependencies=dependencies)
        elif run is False:
            depend(self, sources + dependencies)
        else:
            rule(run, self, sources, dependencies=dependencies)

    def __status__(self, status):

        artefact.__status__(self, status)
        if self.status is True:
            self.outcome = xpass if self.xoutcome == fail else pass_
        else:
            self.outcome = xfail if self.xoutcome == fail else fail
        colour = {pass_: 'green',
                  fail: 'red',
                  xpass: 'yellow',
                  xfail: 'green'}[self.outcome]
        label = {pass_: 'PASS',
                 fail: 'FAIL',
                 xpass: 'XPASS',
                 xfail: 'XFAIL'}[self.outcome]
        logger.info('{}: {}'.format(self.qname, output.coloured(label, colour)))



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..feature import set
from ..artefact import artefact, notfile, nocare
from ..delayed import delayed
import sqlite3
import hashlib
import os
from os.path import exists, join
import logging

logger = logging.getLogger('config')


class cache(object):

    def __init__(self, builddir):
        fdir = join(builddir, '.faber')
        if not exists(fdir):
            os.makedirs(fdir)
        self.filename = join(fdir, 'configcache')
        self.conn = sqlite3.connect(self.filename)
        # Create table if it doesn't exist yet.
        if not next(self.conn.execute('SELECT name FROM sqlite_master '
                                      'WHERE type="table" AND name="checks"'), None):
            self.conn.execute('CREATE TABLE checks '
                              '(key TEXT, status INTEGER, type TEXT, value TEXT)')

    def finish(self):
        if self.conn:
            self.conn.commit()
            self.conn.close()

    def reset(self, level):
        self.conn.commit()
        if level > 1:
            self.conn.execute('DELETE FROM checks')

    def clean(self):
        self.conn.close()
        self.conn = None
        logger.info('cleaning config cache')
        os.remove(self.filename)
        self.conn = sqlite3.connect(self.filename)

    def __contains__(self, key):
        with self.conn:
            a = self.conn.execute('SELECT key FROM checks WHERE key=?', (key,))
        value = next(a, None)
        return bool(value)

    def __setitem__(self, key, value):
        status, result = value[0], value[1]
        with self.conn:
            self.conn.execute('INSERT INTO checks VALUES(?,?,?,?)',
                              (key, status, type(result).__name__, str(result)))

    def __getitem__(self, key):
        with self.conn:
            a = self.conn.execute('SELECT status, type, value FROM checks WHERE key=?',
                                  (key,))
        status, type, value = next(a, (None, None, None))
        value = {'str': lambda x: x,
                 'unicode': lambda x: x,
                 'bool': lambda x: eval(x),
                 'int': lambda x: int(x)}[type](value)
        return status, value


class logfiles(dict):

    def __getitem__(self, m):
        if m not in self:
            self[m] = open(join(m.builddir, 'config.log'), 'a')
        return dict.__getitem__(self, m)

    def clear(self):
        for f in self.values():
            f.close()
        dict.clear(self)

    def clean(self):
        for f in self.values():
            f.close()
            os.remove(f.name)

    def reset(self):
        for f in self.values():
            f.seek(0)
            f.truncate(0)


class check(artefact):
    """A check is an artefact that performs some tests (typically involving compilation),
    then stores the result in a cache, so it doesn't need to be performed again,
    until the cache is explicitly cleared."""

    cache = None
    logfiles = None

    def __init__(self, name, features=(), if_=(), ifnot=()):

        self.result = None
        artefact.__init__(self, name, attrs=notfile|nocare, features=features)
        self.logfile = check.logfiles[self.module]
        # The 'condition' here is simply the value of the check's status member.
        self.use = delayed(lambda: set.instantiate(if_) if self.status else set.instantiate(ifnot), self)
        self.reset()

    def reset(self):
        key = str((self.name, str(self.features))).encode('utf-8')
        self._cache_key = hashlib.md5(key).hexdigest()
        self.cached = self._cache_key in check.cache
        if self.cached:
            self.status, self.result = check.cache[self._cache_key]

    def __status__(self, status):
        logger.debug('check.__status__({})'.format(status))
        if self.cached:
            return  # the cached value takes precedence
        artefact.__status__(self, status)
        if not self.status or self.result is None:
            self.result = self.status
        check.cache[self._cache_key] = (self.status, self.result)



#
# Copyright (c) 2017 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.artefacts import install
from faber.rule import rule
from faber.tools import fileutils
from faber.feature import set
from faber.module import module
from os import mkdir
from os.path import join, exists, splitdrive
import pytest


def relpath(stage='', prefix=''):
    # combine stage with prefix to form the actual path
    if stage:
        if stage[-1] != '/':
            stage += '/'
        prefix = splitdrive(prefix)[1]
    return join(stage + prefix)


@pytest.mark.parametrize('stage, prefix', [('stage', None),
                                           ('stage', 'my/prefix'),
                                           ('stage', '/my/prefix')])
@pytest.mark.usefixtures('module')
def test_installation(stage, prefix):
    """Install a (source) file, directory, as well as a generated
    file, using different stage and prefix settings."""

    srcdir = module.current.srcdir
    filename = 'file'
    with open(join(srcdir, filename), 'w') as o:
        o.write('something')
    dirname = 'directory'
    mkdir(join(srcdir, dirname))

    assert exists(join(srcdir, filename))
    assert exists(join(srcdir, dirname))

    fs = set()
    if prefix is not None:
        fs += install.prefix(prefix, base='')
    if stage is not None:
        stage = join(module.current.builddir, stage)
        fs += install.stage(stage, base='')
    a = install.installed(rule(fileutils.touch, 'empty-file'), 'bin', features=fs)
    b = install.installed(filename, 'src', features=fs)
    c = install.installed(dirname, 'src', features=fs)
    i = install.installation('install', [a, b, c], features=fs)
    assert i.update()
    manifest = [relpath(stage, f.strip())
                for f in open(i.manifest._filename).readlines()]
    print('manifest:', manifest)
    assert len(manifest) == 3
    for f in [a, b, c]:
        print(f._filename)
    for f in manifest:
        assert exists(f)



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

multi= 0x0001
path= 0x0002
incidental= 0x0004

from .value import value  # noqa F401
from .feature import feature  # noqa F401
from .set import set, lazy_set  # noqa F401
from .map import map, translate, join, select_if  # noqa F401



#
# Copyright (c) 2017 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..action import action
from ..tool import tool
import shutil


class archive(action):

    def __init__(self, format=None):
        action.__init__(self)
        if format:
            self.format = format
        elif shutil._BZ2_SUPPORTED:
            self.format = 'bztar'
        elif shutil._ZLIB_SUPPORTED:
            self.format = 'gztar'
        else:
            self.format = 'tar'

    def command(self, targets, sources):
        archive = targets[0]
        installation = sources[0]
        stage = str(installation.features.stage)
        shutil.make_archive(archive.name, self.format, stage)


class archiver(tool):

    @staticmethod
    def extension(format):
        if format == 'bztar':
            return '.tar.bz2'
        elif format == 'gztar':
            return '.tar.gz'
        elif format == 'tar':
            return '.tar'
        elif format == 'zip':
            return '.zip'
        else:
            raise ValueError('unsupported archive format {}'.format(format))

    archive = archive



#
# Copyright (c) 2018 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..action import action
from ..feature import set
from ..tool import tool
import subprocess
import re


def validate(cls, command, version, features):

    features = set.instantiate(features)
    version = version or cls.find_version_requirement(features)
    v = subprocess.check_output([command, '--version']).decode().strip()
    v = re.match('.* ([0-9.]+)', v).group(1)
    if version and v != version:
        raise ValueError('{} version mismatch: expected {}, got {}'
                         .format(command, version, v))
    else:
        version = v
    return command, version, features


class sphinx(tool):

    html = action('sphinx-build -b html $(>) $(<)')

    def __init__(self, name='sphinx', command=None, version='', features=()):

        command, version, features = validate(self.__class__, command or 'sphinx-build',
                                              version, features)
        tool.__init__(self, name=name, version=version)
        self.features |= features
        if command:
            self.html.subst('sphinx-build', command)



#
# Copyright (c) 2017 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.feature import feature, multi, path, incidental, set
from faber.artefact import artefact
from faber.artefact import notfile, always
from faber.rule import rule, depend
from faber import scheduler
from test.common import pyecho
import pytest
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


class placeholder(artefact):
    # a placeholder takes the place of a conditional dependency until
    # the condition is evaluated and the substition made (or not)
    def __init__(self, d, p, c):
        artefact.__init__(self, p.name, features=p.features, attrs=notfile)
        self.prereq = p
        self.dep = d
        self.cond = c

    def __status__(self, status):
        artefact.__status__(self, status)
        self.prereq.features.eval(update=False)
        if self.cond(self.prereq.features):
            depend(self.dep, [self.prereq])


@pytest.mark.usefixtures('module')
def test_conditional_dependency():
    """Test the workflow of conditional dependencies"""

    name = feature('name', attributes=multi|path|incidental)
    # artefact to compute the condition
    a = rule(None, 'a', attrs=notfile|always)
    fs = set(name(a.filename))
    # conditional artefact
    b1 = rule(pyecho, 'b1', features=fs, attrs=notfile|always)
    b2 = rule(pyecho, 'b2', features=fs, attrs=notfile|always)
    # c gets built unconditionally
    # but will include b1 or b2 only if condition is met
    c = artefact('c', attrs=notfile)
    c = rule(pyecho, c, dependencies=[placeholder(c, b1, set.name == ''),
                                      placeholder(c, b2, set.name != '')])
    with patch('faber.action.action.__status__') as recipe:
        scheduler.update(c)
        output = [i[0][4].strip() for i in recipe.call_args_list]
        assert 'b1 <-' not in output and 'b2 <-' in output



#
# Copyright (c) 2017 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..artefact import source
import yaml


class node(object):
    """Convert a (nested) dictionary into an ordinary object."""
    def __init__(self, d):
        for a, b in d.items():
            if isinstance(b, (list, tuple)):
                setattr(self, a, [node(x) if isinstance(x, dict) else x for x in b])
            else:
                setattr(self, a, node(b) if isinstance(b, dict) else b)


def load_info(filename):
    """Load a package metadata file."""
    with open(filename) as f:
        data = yaml.safe_load(f)
        return node(data)


class info(source):

    def __init__(self, name):
        source.__init__(self, name)
        self.doc = load_info(self._filename)



#
# Copyright (c) 2020 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.project import project, buildinfo
from faber.module import module as M
from os.path import join
from test.common import tempdir
from os import mkdir
import pytest


@pytest.fixture(autouse=True)
def module():
    with tempdir() as root:
        srcdir = join(root, 'test-source')
        mkdir(srcdir)
        builddir = join(root, 'test-build')
        info = buildinfo(builddir, srcdir)
        with project(info):
            with M('test', srcdir, builddir, process=False):
                yield



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from __future__ import absolute_import
from . import scheduler
from .action import action
from .feature import set
from .feature.condition import expr as fexpr
from .artefact import artefact, source, conditional, notfile
from .utils import aslist
from types import FunctionType, MethodType
import logging

logger = logging.getLogger('rules')
feature_logger = logging.getLogger('features')


def conditional_dependency(a, d):
    if d.condition is None:
        # unconditional
        return d
    elif not isinstance(d.condition, fexpr):
        # eval using builtin __bool__
        return d if d.condition else None
    elif not d.features.dependencies():
        # eval using feature.condition
        result = d.condition(d.features.eval(update=False))
        feature_logger.info('condition "{}" yields {}'.format(d.condition, result))
        return d if result else None
    else:
        # postpone eval using feature.condition
        feature_logger.info('postponing evaluation of condition "{}"'
                            .format(d.condition))
        if not isinstance(d, conditional):
            d = conditional(d, d.condition)
        d.dependent.append(a)
        return d


def depend(target, dependencies):
    """Declare `target` to depend on `dependencies`."""

    dependencies = aslist(dependencies)
    # wrap conditional dependencies
    dependencies = list(filter(None, [conditional_dependency(target, d)
                                      for d in dependencies]))
    scheduler.add_dependency(target, dependencies)


def _rule(recipe, targets, sources, deps, attrs, features, module, path_spec, logfile):

    targets = aslist(targets)
    sources = aslist(sources)
    logger.info('rule: {} <- {} with {}'.format(targets, sources, recipe))
    if not features and isinstance(targets[0], artefact):
        features = targets[0].features.copy()
    else:
        features = module.features | set.instantiate(features)

    if recipe:
        # instantiate recipe
        if type(recipe) in (FunctionType, MethodType):
            recipe = action(recipe.__name__, recipe)
        elif recipe and recipe.abstract:
            # look up an appropriate tool providing the action
            recipe = recipe.instantiate(features.essentials())
        features |= recipe.features
        path_spec = recipe.path_spec if not path_spec else path_spec

    deps = aslist(deps)
    # instantiate artefacts for sources and dependencies
    sources = [source.instantiate(s, module) for s in sources]
    deps = [artefact.instantiate(d, module) for d in deps]

    # instantiate artefacts for targets
    def instantiate(a):
        if isinstance(a, artefact):
            a.attrs |= attrs
            a.features |= features
            a.features |= artefact.combine_use(sources)
            a.path_spec = path_spec if path_spec else a.path_spec
        else:
            a = artefact(a, attrs, features=features, path_spec=path_spec,
                         module=module, logfile=logfile)
        return a
    targets = [instantiate(t) for t in targets]
    deps = deps + sources
    if deps:
        for t in targets:
            depend(t, deps)
    if recipe:
        recipe.submit(targets, sources)

    return targets


def rule(recipe, targets, sources=[], dependencies=[],
         attrs=0, features=(), module=None, path_spec='', logfile=None):
    """Express how to generate `artefacts`. In the simplest case this involves a
    source and a recipe...

    Arguments:
      * recipe: an action taking `sources` as input and producing `targets`.
      * targets: what to make. This can be a string (typically a filename),
                 an existing `artefact` object, or a list of either strings or
                 artefacts (or a mix thereof).

      * sources: the source to make the artefact from. As with `artefacts`, this
                 may be a string, a artefact, or a list of those.

      * depends: additional dependencies not to be used as source.
      * attrs:  flags
      * features: a set of features
      * module: the module to instantiate the artefact in"""

    from .module import module as M
    module = module or M.current
    targets = _rule(recipe, targets, sources, dependencies, attrs, features, module,
                    path_spec, logfile)
    return targets[0] if len(targets) == 1 else targets


def alias(target, source, attrs=notfile, module=None):
    """Declare `target` to be an alias for `source`, i.e., update `source` whenever
    updating `target` is requested."""
    return rule(None, target, dependencies=source, attrs=attrs|notfile, module=module)



#
# Copyright (c) 2018 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.scheduler.artefact import *
from faber.scheduler.recipe import recipe
import os


class frontend(object):
    """Emulate a faber.artefact enough to support testing the backend."""

    class fset(object):
        def eval(self, update=True): pass

    def __init__(self, name, attrs):
        self.name = name
        self.boundname = self.name
        self.attrs = attrs
        self.features = frontend.fset()
        self.logfile=False

    def __status__(self, status): pass


def touch(fname, times=None):
    with open(fname, 'a'):
        os.utime(fname, times)


class action(object):

    def __init__(self):
        self.command = lambda targets, sources, **vars: [touch(t.boundname) for t in targets]

    def map(self, features): return {}
    def __status__(self, *args): pass


def make_artefact(name, attrs=0, touch=False, prerequisites=[]):
    a = artefact(frontend(name, attrs), prerequisites=prerequisites)
    a.recipe = recipe(action(), [a], []) if touch else None
    return a



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from . import multi
import os


class map(object):
    """A map maps a feature-set to variables."""

    def __init__(self, feature, func=None, *args, **kwds):
        """Create a map from a function that takes a feature value and returns a list of strings.
        The default function returns the feature value as-is."""

        if func is None:
            func = lambda fv: fv._value if fv is not None else []

        def wrapper(fs):
            fv = next(iter(filter(lambda v: v._type is feature, fs.values())), None)
            return func(fv, *args, **kwds)
        self._funcs = [wrapper]

    def __iadd__(self, other):
        self._funcs += other._funcs
        return self

    def __add__(self, other):
        self._funcs += other._funcs
        return self

    def __call__(self, fs):
        return ' '.join([i for func in self._funcs for i in func(fs)])


def translate(fv, prefix='', suffix=''):
    """Map a multi feature to a list of strings, using a prefix and suffix."""

    if not fv:
        return []
    elif fv._type.attributes & multi:
        return ['{}{}{}'.format(prefix, v, suffix) for v in fv._value] if fv else []
    else:
        return ['{}{}{}'.format(prefix, fv._value, suffix)] if fv else []


def join(fv, char=os.pathsep):

    return [char.join([v for v in fv._value] if fv else [])]


def select_if(fv, ref, value):
    """Return a given value if a given feature has a specific value"""

    return [value] if fv and fv._value == ref else []



#
# Copyright (c) 2017 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.artefact import notfile, always
from faber.rule import rule
from faber.tools import fileutils
from os.path import exists
import pytest
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


@pytest.mark.usefixtures('module')
def test_touch():
    """Check that 'touch' and 'rm' actions work across platforms."""

    a = rule(fileutils.touch, 'empty-file')
    with patch('faber.action.action.__status__'):
        assert a.update()
        assert exists(a._filename)


@pytest.mark.usefixtures('module')
def test_copy():
    """Check that 'touch' and 'rm' actions work across platforms."""

    a = rule(fileutils.touch, 'empty-file')
    b = rule(fileutils.copy, 'clone', a)
    with patch('faber.action.action.__status__'):
        assert b.update()
        assert exists(b._filename)


@pytest.mark.usefixtures('module')
def test_remove():
    """Check that 'touch' and 'rm' actions work across platforms."""

    a = rule(fileutils.touch, 'empty-file')
    b = rule(fileutils.remove, 'cleanup', a, attrs=notfile|always)
    with patch('faber.action.action.__status__'):
        assert a.update()
        assert exists(a._filename)
        assert b.update()
        assert not exists(a._filename)



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..artefact import artefact, notfile, always
from ..rule import rule
from .. import output
from .. import logging
from .suite import suite
from . import pass_, fail, xfail


class report(artefact):
    """A test report runs a set of tests and reports a summary."""

    def __init__(self, name, tests, fail_on_failures=False):
        """Construct a report. Arguments are the tests to be performed."""

        artefact.__init__(self, name, attrs=notfile|always)
        rule(self.print_summary, self, tests)
        self._tests = tests
        self.fail_on_failures = fail_on_failures

    @property
    def tests(self):
        for t in self._tests:
            if isinstance(t, suite):
                # Use `yield from` once we stop supporting Python 2.7
                for i in t:
                    yield i
            else:
                yield t

    def print_summary(self, targets, sources):
        """Print a summary of the report."""

        logger = logging.getLogger('summary')

        passes = sum(1 for s in self.tests if s.outcome == pass_)
        failures = [s for s in self.tests if s.outcome == fail]
        xfailures = sum(1 for s in self.tests if s.outcome == xfail)
        skipped = sum(1 for s in self.tests if s.outcome is None)

        p = passes and '{} pass'.format(passes) or ''
        if passes > 1: p += 'es'
        nfailures = len(failures)
        colour = 'red' if nfailures else 'green'
        f = nfailures and '{} failure'.format(nfailures) or ''
        if nfailures > 1: f += 's'
        x = xfailures and '{} expected failure'.format(xfailures) or ''
        if xfailures > 1: x += 's'
        s = skipped and '{} skipped'.format(skipped) or ''
        line = ', '.join([o for o in [p, f, x, s] if o])

        logger.info(output.coloured('test summary: ' + line, colour, attrs=['bold']))
        if failures:
            logger.info(output.coloured('failures:', 'red', attrs=['bold']))
            for f in failures:
                logger.info(output.coloured(f.qname, 'red', attrs=['bold']))
                if f.output[1]:  # if there is output on stderr...
                    logger.info(f.output[1])
            if self.fail_on_failures:
                return False
        return True



#
# Copyright (c) 2017 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.artefact import artefact
from faber.artefact import notfile, always, intermediate
from faber.rule import rule, depend
from faber.tools import fileutils
from faber import scheduler
from os.path import exists
import pytest
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


@pytest.mark.usefixtures('module')
def test_composite():
    """Test the workflow of faber.artefacts.binary"""

    src = rule(fileutils.touch, 'src')
    bin = artefact('bin')

    def assemble(targets, sources):
        obj = rule(fileutils.copy, 'obj', src,
                   attrs=intermediate, module=targets[0].module)
        rule(fileutils.copy, bin, obj,
             attrs=intermediate, module=targets[0].module)

    def test(targets, sources):
        print('testing {}'.format(sources[0].name))
        assert exists(sources[0]._filename)

    # make a dependency graph
    ass = rule(assemble, 'ass', attrs=notfile|always)
    # make a binary
    depend(bin, ass)
    # and test it
    test = rule(test, 'test', bin, attrs=notfile)

    with patch('faber.action.action.__status__') as recipe:
        scheduler.update([test])
        (_, status, _, _, output, _), kwds = recipe.call_args_list[-1]
        assert output == 'testing bin\n'
        assert status is True



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from .action import action  # noqa F401
from .tool import tool  # noqa F401
from .artefact import *  # noqa F401
from .rule import rule, alias, depend  # noqa F401
from .module import module  # noqa F401
from . import scheduler  # noqa F401



#
# Copyright (c) 2017 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.artefact import artefact
from faber.artefact import notfile, always, intermediate
from faber.rule import rule, depend, alias
from faber.tools import fileutils
from faber import scheduler
import pytest
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


@pytest.mark.usefixtures('module')
def test_composite():
    """Test the workflow of faber.config.try_compile"""

    src = rule(fileutils.touch, 'src', attrs=intermediate|always)
    obj = artefact('obj', attrs=intermediate)
    check = artefact('check', attrs=notfile)

    def assemble(targets, sources):
        rule(fileutils.copy, obj, src,
             attrs=intermediate, module=targets[0].module)

    # make a dependency graph
    ass = rule(assemble, 'ass', attrs=notfile|always)
    # make a binary
    depend(obj, dependencies=ass)
    # and test it
    check = alias(check, obj)

    with patch('faber.action.action.__status__') as recipe:
        scheduler.update([check])
        (_, status, _, _, _, _), kwds = recipe.call_args_list[-1]
        assert recipe.call_count == 3
        assert status is True



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.artefact import notfile, intermediate
from faber.rule import rule
from faber.tools import fileutils
from faber import scheduler
from test.common import pyecho
from os.path import exists
import pytest


@pytest.mark.usefixtures('module')
def test_intermediate():
    """This is a reduced version of the config example.
    It shows a bug in bjam's state engine. The test passes
    if all files are non-intermediate."""

    c = rule(fileutils.touch, 'c', attrs=intermediate)
    d = rule(fileutils.touch, 'd', c)
    o = rule(fileutils.touch, 'o', [c, d], attrs=intermediate)
    b = rule(fileutils.touch, 'b', o, attrs=intermediate)
    t = rule(pyecho, 't', b, attrs=notfile)
    assert scheduler.update(t)
    assert exists(b._filename)



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from . import multi, path
from operator import iadd
from functools import partial
import os.path
import logging

logger = logging.getLogger('features')


class feature(object):

    _registry = dict()

    def _constr(self, v):
        from .value import value
        if type(v) is value:
            return v
        elif type(v) is tuple:
            return value(self, v, None)
        else:
            return value(self, v, None)

    def _register(self):
        feature._registry[self.name] = self
        logger.info('defining feature "{}"'.format(self.name))

    @staticmethod
    def lookup(name):
        return feature._registry[name]

    def __new__(cls, *args, **kwds):
        # if any of the keyword arguments are features, construct a compound-feature
        composite = any([isinstance(a, feature) for a in args])
        if composite:
            cls = composite_feature
        inst = object.__new__(cls)
        return inst

    def __init__(self, name, values=None, attributes=0, sub=False):
        """Create a feature."""

        self.name = name
        self.values = values
        self.attributes = attributes
        if not sub:
            self._register()

    def __call__(self, *args, **kwds):
        """Instantiate a feature value."""
        from .value import value
        from ..delayed import delayed
        cond = kwds.pop('condition', None)
        # HACK: we need to define what argument types to allow
        if args and isinstance(args[0], delayed):
            # simply resubmit the value in the delayed's `eval`
            return args[0].apply(partial(self.__call__, **kwds))
        else:
            return value(self, self._validate(*args, **kwds), cond)

    def _join(self, values):
        # create a new value by joining the given arguments
        if self.attributes & multi:
            kwds = {'base': ''} if self.attributes & path else {}
            return self(*tuple([v for value in values for v in value._value]), **kwds)
        else:
            return values[0]  # only one is allowed. Assuming all are identical, return the first.

    def _cassign(self, op, v1, v2):

        from .value import value
        assert v1._type is self
        t2 = v2._type if isinstance(v2, value) else self
        if self is not t2:
            raise ValueError('cannot assign values from different features "{}" and "{}"'
                             .format(self.name, t2.name))
        v2 = v2._value if isinstance(v2, value) else v2
        if v2 is not None:
            if self.attributes & multi:
                if op is iadd:
                    v1._value += v2
                else:
                    v1._value += tuple(v for v in v2 if v not in v1._value)
            else:
                if v1._value is not None and v1._value != v2:
                    raise ValueError('can not override {} by {}'.format(v1._value, v2))
                v1._value = v2
        return v1

    def _validate(self, *args, **kwds):
        def adjust_path(path, base):
            # At this point 'base' may only be None if there is no current module
            if base is None and not os.path.isabs(path):
                raise ValueError('relative path "{}" in feature "{}" may not be defined outside module'
                                 .format(path, self.name))
            else:
                return os.path.normpath(os.path.join(base or '', path))

        if self.attributes & multi:
            if self.values is not None:
                for a in args:
                    if a not in self.values:
                        raise ValueError('invalid value "{}" in feature "{}"'
                                         .format(a, self.name))
                value = args
            elif self.attributes & path:
                base = kwds.get('base', None)
                if base is None:
                    from ..module import module
                    base = module.current.srcdir if module.current else None
                value = [adjust_path(a, base) for a in args]
            else:
                value = args[:]
        elif len(args) > 1:
            raise ValueError('feature "{}" requires a single value'.format(self.name))
        elif args:
            if self.values is not None and args[0] not in self.values:
                raise ValueError('invalid value "{}" in feature "{}"'
                                 .format(args[0], self.name))
            elif self.attributes & path:
                base = kwds.get('base', None)
                if base is None:
                    from ..module import module
                    base = module.current.srcdir if module.current else None
                value = adjust_path(args[0], base)
            else:
                value = args[0]
        else:
            value = None
        return value


class composite_feature(feature):

    def __init__(self, name, *args, **kwds):
        """Create a composite feature."""

        sub = kwds.pop('sub', False)
        assert not kwds
        self.name = name
        self.subfeatures = []
        # ...then handle the keyword arguments as subfeatures.
        for s in args:
            assert isinstance(s, feature)
            self.subfeatures.append(s)
        if not sub:
            self._register()

    def __call__(self, *args, **kwds):
        """Instantiate a feature value."""
        from .value import composite_value
        if args:
            assert len(args) == 1 and not kwds
            sf = args[0].split('-')
            values = {s.name: sf[i] for i, s in enumerate(self.subfeatures)}
        else:
            values = {s.name: None for s in self.subfeatures}
            values.update({s.name: s._constr(kwds.get(s.name)) for s in self.subfeatures})
        return composite_value(self, **values)

    def _join(self, values):
        # create a new value by joining the given arguments
        return self(**{s.name: s._join([getattr(v, s.name) for v in values])
                       for s in self.subfeatures})

    def _cassign(self, op, v1, v2):

        from .value import value
        assert v1._type is self
        t2 = v2._type if isinstance(v2, value) else self
        if self is not t2:
            raise ValueError('cannot join values from different features "{}" and "{}"'
                             .format(self.name, t2.name))
        assert isinstance(v2, value)
        for s in self.subfeatures:
            s._cassign(op, getattr(v1, s.name), getattr(v2, s.name))
        return v1



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from .try_compile import try_compile
from .. import types


class has_cxx11(try_compile):

    src = r"""#if __cplusplus < 201103L
#error no C++11
#endif"""

    def __init__(self, features=(), if_=(), ifnot=()):
        try_compile.__init__(self, 'has_cxx11', has_cxx11.src, types.cxx, features,
                             if_, ifnot)


class has_cxx14(try_compile):

    src = r"""#if __cplusplus < 201402L
#error no C++14
#endif"""

    def __init__(self, features=(), if_=(), ifnot=()):
        try_compile.__init__(self, 'has_cxx14', has_cxx14.src, types.cxx, features,
                             if_, ifnot)


class has_cxx17(try_compile):

    src = r"""#if __cplusplus < 201500L
#error no C++17
#endif"""

    def __init__(self, features=(), if_=(), ifnot=()):
        try_compile.__init__(self, 'has_cxx17', has_cxx17.src, types.cxx, features,
                             if_, ifnot)



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from __future__ import absolute_import
from . import types
from .feature import set
from .delayed import delayed_property
from .utils import path_formatter, add_metaclass
from . import logging
from os.path import normpath, join
from collections import defaultdict
from functools import reduce

scheduler_logger = logging.getLogger('scheduler')
feature_logger = logging.getLogger('features')

intermediate= 0x0001
nocare= 0x0002
notfile= 0x0004
always= 0x0008
leaves= 0x0010
noupdate= 0x0020
rmold= 0x0080
internal=0x0100
precious= 0x0200
nopropagate = 0x0400


class artefact(object):
    """An artefact can be generated by a build."""

    _path_formatter = path_formatter()
    _qnames = defaultdict(list)

    @staticmethod
    def finish():
        artefact._qnames.clear()

    @staticmethod
    def iter():
        """Iterate over all registered artefacts."""

        return iter(a for qname in artefact._qnames.values() for a in qname)

    @staticmethod
    def lookup(qname):
        """Find an artefact by (qualified) name."""

        if qname not in artefact._qnames:
            raise KeyError(qname)
        else:
            return artefact._qnames[qname]

    @classmethod
    def instantiate(cls, a, module=None):
        return a if isinstance(a, artefact) else cls(a, module=module)

    @staticmethod
    def combine_use(artefacts):
        """return the union of the use features from all artefacts"""
        return reduce(lambda fs, a: fs|a.use, artefacts, set())

    @property
    def qname(self):
        """The qualified name of this artefact."""
        return self.module.qname(self.name)

    @property
    def id(self):
        """Yield a unique identifier for this artefact."""
        return '%s-%x' % (self.qname, id(self))

    @property
    def isfile(self):
        """Just a convenience property for a frequent query"""
        return not self.attrs & notfile

    @property
    def relpath(self):
        """The relative path corresponding to the artefact's feature set."""
        f = artefact._path_formatter
        return normpath(f.format(self.path_spec, **self.features))

    @property
    def _filename(self):
        """The (full) filename of this artefact, if it represents a file."""
        return normpath(join(self.module.builddir, self.relpath, self.name))

    @delayed_property
    def filename(self):
        return self._filename

    @property
    def boundname(self):
        """Return either self.filename or self.qname depending on whether
        this artefact represents a file."""
        return self._filename if self.isfile else self.qname

    def __init__(self, name, attrs=0, type=None, module=None,
                 features=(), use=(), condition=None,
                 path_spec='',
                 logfile=None):
        self.name = name
        from .module import module as M
        self.module = module or M.current
        self.attrs = attrs
        self.path_spec = path_spec
        # Make sure sources are indeed feature-less (and free of dependencies !)
        if isinstance(self, source):
            assert not features
            self.features = set()
        else:
            self.features = self.module.features | set.instantiate(features)
        self.use = set.instantiate(use)
        self.condition = condition
        self.status = None
        self.type = type
        if not self.type and self.isfile:
            # just a guess, subclasses may override
            self.type = types.type.discover(self._filename)
        self.logfile = logfile
        self._register()

    def __call__(self, features):
        """Clone this artefact, and apply new features."""

        import copy
        clone = copy.copy(self)
        clone.features = self.features.copy()
        clone.features.update(set.instantiate(features))
        clone._register()
        return clone

    def update(self):
        from . import scheduler
        return scheduler.update(self)

    def reset(self):
        self.status = None

    def _register(self):
        from . import scheduler
        artefact._qnames[self.qname].append(self)
        scheduler.define_artefact(self)
        deps = self.features.dependencies()
        if deps:
            scheduler.add_dependency(self, deps)

    def __str__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.qname)

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.id)

    def __status__(self, status):
        self.status = status


class source_type(type):
    def __call__(cls, name, *args, **kwds):
        """Make sure there is only one instance per source (file)."""
        from .module import module as M
        module = kwds.get('module') or M.current
        qname = module.qname(name)
        if qname not in cls._instances:
            cls._instances[qname] = type.__call__(cls, name, *args, **kwds)
        return cls._instances[qname]


@add_metaclass(source_type)
class source(artefact):
    """A source is a simple feature-less artefact representing an existing
    file outside faber's control."""

    _instances = {}

    def _register(self):
        # don't depend on anything
        from . import scheduler
        artefact._qnames[self.qname].append(self)
        scheduler.define_artefact(self, bind=True)

    @property
    def _filename(self):
        return join(self.module.srcdir, self.name)


class conditional(artefact):

    def __init__(self, a, condition, features=()):
        artefact.__init__(self, 'c:' + a.name, features=features,
                          attrs=notfile, condition=condition, module=a.module)
        self.a = a
        self.dependent = []  # set by depend()

    def reset(self):
        # do not reset this artefact as we can't undo its action
        pass

    def __status__(self, status):
        artefact.__status__(self, status)
        result = self.condition(self.features.eval(update=False))
        feature_logger.info('condition "{}" yields {}'
                            .format(self.condition, result))
        if result:
            from . import scheduler
            for d in self.dependent:
                scheduler.add_dependency(d, self.a)


def init():
    """reset globals"""

    source._instances.clear()



#
# Copyright (c) 2017 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..action import action
from ..platform import os

if os == 'Windows':

    touch = action('touch', """
@echo off
    setlocal enableextensions disabledelayedexpansion

    (for %%a in ($(<)) do if exist "%%~a" (
        pushd "%%~dpa" && ( copy /b "%%~nxa"+,, & popd )
    ) else (
        type nul > "%%~fa"
    )) >nul 2>&1""")

    copy = action('copy', 'copy /b $(>) $(<)')
    remove = action('remove', 'del $(>)')

else:
    touch = action('touch', 'touch $(<)')
    copy = action('copy', 'cp -r $(>) $(<)')
    remove = action('remove', 'rm -rf $(>)')



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from . import multi, incidental
from operator import iadd, ior
import copy


class value(object):

    def __init__(self, type, v, condition, **kwds):
        """Construct a feature value.
        Arguments:

          * type: the feature
          * v: the value (list if this is a 'multi' feature, a single value otherwise)
          * condition: a condition in case this is a conditional feature
          * kwds: subfeatures, if this is a composite feature value"""
        self._type = type
        self._name = type.name
        self._value = v
        assert not isinstance(v, value)
        self._condition = condition

    def __bool__(self):
        return bool(self._value)

    __nonzero__ = __bool__  # python 2

    def __copy__(self):
        v = copy.copy(self._value)
        return value(self._type, v, self._condition)

    def copy(self):
        return self.__copy__()

    def __call__(self, condition):
        """Create a conditionalized version of self."""
        clone = self.copy()
        clone._condition = condition
        return clone

    def __iadd__(self, other):
        return self._type._cassign(iadd, self, other)

    def __ior__(self, other):
        return self._type._cassign(ior, self, other)

    def __eq__(self, other):
        if type(other) is value:
            if self._type != other._type:
                return False
            return self._value == other._value
        else:
            # allow 'other' to be raw data
            if self._type.attributes & multi:
                return other == self._value or other in self._value
            else:
                return self._value == other

    def __ne__(self, other):
        return not self == other

    def __contains__(self, v):
        return v in self._value

    def matches(self, other):
        """Report whether 'other' matches 'self', when used as a spec.
        For incidental features this unconditionally returns True.
        Otherwise this reports true iff all defined values in self equal
        the corresponding values in 'other'."""
        if isinstance(other, value):
            assert self._type == other._type

        if self._type.attributes & incidental:
            return True
        else:
            if not isinstance(other, value):
                return other in self._value if self._type.attributes & multi else other == self._value
            else:
                return (self._value is None or
                        other._value is None or
                        self._value == other._value)

    def __repr__(self):
        value = self._value
        return '<value {}={}>'.format(self._type.name, value)

    def __str__(self):
        if self._type.attributes & multi:
            return ' '.join(str(v) for v in self._value)
        else:
            return str(self._value)

    def serialize(self, prefix=''):
        if prefix: prefix += '.'
        return {prefix + self._type.name: self._value}

    def eval(self, set):

        if self._condition is not None and self._condition(set):
            clone = self.copy()
            clone._condition = None
            return (clone,)
        else:
            return ()


class composite_value(value):

    def __init__(self, type, **kwds):
        """Construct a feature value.
        Arguments:

          * type: the feature
          * kwds: subfeatures"""
        self._type = type
        self._name = type.name
        self.__dict__.update({s.name: kwds.get(s.name) for s in self._type.subfeatures})
        self._condition = None  # FIXME

    def __bool__(self):
        return any([bool(self.__dict__[s.name]) for s in self._type.subfeatures])

    __nonzero__ = __bool__  # python 2

    def __copy__(self):
        # make sure to copy all values
        kwds = {s.name: self.__dict__[s.name].copy() for s in self._type.subfeatures
                if self.__dict__[s.name] is not None}
        return composite_value(self._type, **kwds)

    def __eq__(self, other):
        if isinstance(other, composite_value):
            for s in self._type.subfeatures:
                if getattr(self, s.name) != getattr(other, s.name):
                    return False
            return True
        else:
            return False

    def __contains__(self, v):
        return False

    def matches(self, other):
        """Report whether 'other' matches 'self', when used as a spec.
        For incidental features this unconditionally returns True.
        Otherwise this reports true iff all defined values in self equal
        the corresponding values in 'other'."""
        if not isinstance(other, composite_value):
            return False  # comparison to raw data is only allows for non-compound values.
        elif self._type != other._type:
            return False
        for s in self._type.subfeatures:
            if not getattr(other, s.name).matches(getattr(self, s.name)):
                return False
        return True

    def __repr__(self):
        value = '<...>'
        return '<value {}={}>'.format(self._type.name, value)

    def __str__(self):
        values = [getattr(self, s.name) for s in self._type.subfeatures]
        return '<{}>'.format('.'.join([str(v) for v in values]))

    def serialize(self, prefix=''):
        if prefix: prefix += '.'
        d = {}
        for s in self._type.subfeatures:
            sf = getattr(self, s.name)
            d.update(sf.serialize(prefix + self._type.name) if sf else {})
        return d

    def eval(self, set):

        return ()



#
# Copyright (c) 2017 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber import cli
from test.common import tempdir, argv
import pytest
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


def test_cli_mixed_args():
    # Mixing positional and non-positional arguments is tricky
    # (see https://bugs.python.org/issue14191) and requires a
    # workaround. Make sure it works as expected...

    with patch('faber.project.project.build'):
        with tempdir() as b:
            build = '--builddir={}'.format(b)
            with argv(['faber', build, 'a', 'a=b', '-s']):
                assert cli.main()
        with tempdir() as b:
            build = '--builddir={}'.format(b)
            with argv(['faber', build, 'a', '-s', 'a=b']):
                assert cli.main()
        with tempdir() as b:
            build = '--builddir={}'.format(b)
            with argv(['faber', build, 'a', '-s', 'a=b', '--with-c=e']):
                assert cli.main()
        with tempdir() as b, pytest.raises(SystemExit):
            build = '--builddir={}'.format(b)
            with argv(['faber', build, 'a', '-s', 'a=b', '--invalid']):
                cli.main()



#
# Copyright (c) 2019 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.project import *
from test.common import tempdir, write_fabscript
import pytest
from os.path import join, exists
from os import mkdir


def test_no_args():

    with tempdir():
        info = buildinfo(None)
        assert info.srcdir is None
        assert info.builddir is None


def test_new_project():

    with tempdir() as root:
        # setup
        srcdir = join(root, 'test-source')
        mkdir(srcdir)
        write_fabscript(srcdir, '')
        builddir = join(root, 'test-build')

        # test
        info = buildinfo(builddir, srcdir)
        assert info.srcdir == srcdir
        assert info.builddir == builddir
        info.store()
        assert exists(join(builddir, '.faber', 'info'))


def test_existing_project():

    with tempdir() as root:
        # setup
        srcdir = join(root, 'test-source')
        mkdir(srcdir)
        write_fabscript(srcdir, '')
        builddir = join(root, 'test-build')
        info = buildinfo(builddir, srcdir)
        info.store()

        # test
        info = buildinfo(builddir, srcdir)
        assert info.srcdir == srcdir
        assert info.builddir == builddir


def test_parameters():

    with tempdir() as root:
        # setup
        srcdir = join(root, 'test-source')
        mkdir(srcdir)
        write_fabscript(srcdir, '')
        builddir = join(root, 'test-build')
        info = buildinfo(builddir, srcdir)
        info.parameters = dict(answer='42')
        info.store()

        # test
        info = buildinfo(builddir, srcdir)
        assert info.parameters == dict(answer='42')


def test_inplace_project():

    with tempdir() as root:
        # setup
        srcdir = join(root, 'test-source')
        mkdir(srcdir)
        write_fabscript(srcdir, '')

        # test
        info = buildinfo(srcdir)
        assert info.srcdir == srcdir
        assert info.builddir == srcdir


def test_invalid_project():

    with tempdir() as root:
        # setup
        srcdir = join(root, 'test-source')
        mkdir(srcdir)
        write_fabscript(srcdir, '')
        builddir = join(root, 'test-build')
        info = buildinfo(builddir, srcdir)
        info.store()

        # test
        with pytest.raises(Exception):
            info = buildinfo(builddir, srcdir + '-other')



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.module import module
from faber.error import ScriptError
from test.common import tempdir, write_fabscript
import pytest


@pytest.mark.usefixtures('module')
def test_no_feature():

    script="""
from faber.feature import set
from faber.tools.compiler import include

fs = set()
d = fs.define"""

    with tempdir() as dirpath:
        write_fabscript(dirpath, script)
        with pytest.raises(ScriptError) as e:
            m = module(dirpath)  # noqa F841
        assert e.value.lineno == 6 and 'no feature' in e.value.message


@pytest.mark.usefixtures('module')
def test_no_compiler():

    script="""
from faber.types import cxx
from faber.tools.compiler import target
from faber.config.try_compile import try_compile
features |= target(arch='impossible')
src='int main() {}'
t=try_compile('test', src, cxx, features=features)
default=t
"""

    with tempdir() as dirpath:
        write_fabscript(dirpath, script)
        with pytest.raises(ScriptError) as e:
            m = module(dirpath)  # noqa F841
        assert 'no C++ compiler found' in e.value.message
        # close open files so we can remove dirpath
        from faber import config
        config.finish()



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from .try_run import check_output
from .. import types


class sizeof(check_output):

    src = """#include <stdio.h>
int main(){{ printf("%i", sizeof({}));}}"""

    def __init__(self, c_type, lang_type=types.c, features=()):
        check_output.__init__(self, 'sizeof_' + c_type, sizeof.src.format(c_type), lang_type,
                              features)

    def post_process(self, output):
        self.result = int(output)



#!/usr/bin/env python
#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from setuptools import setup, find_packages
from distutils.command import build, install_data, sdist
import versioneer
import sys
import os
import subprocess

if sys.version_info < (3, 6, 0):
    sys.exit('Faber requires Python 3.6 or newer.')

# allow the in-place import of the version
sys.path.insert(0, 'src')
from faber import cli  # noqa E402

version = versioneer.get_version()

data = [('share/doc/faber-{}'.format(version), ('LICENSE', 'README.md'))]


class build_doc(build.build):

    description = "build documentation"

    def run(self):

        self.announce('building documentation')
        orig = sys.argv
        sys.argv = ['faber', '--srcdir=doc', '--builddir=doc']
        try: cli.main()
        finally: sys.argv = orig


docs = []
if os.path.exists('doc/html'):
    for root, dirs, files in os.walk('doc/html'):
        dest = root.replace('doc/html', 'share/doc/faber-{}'.format(version))
        docs.append((dest,
                    [os.path.join(root, file) for file in files
                     if os.path.isfile(os.path.join(root, file))]))


setup(name='faber',
      python_requires='>=3.6',
      version=version,
      author='Stefan Seefeld',
      author_email='stefan@seefeld.name',
      maintainer='Stefan Seefeld',
      maintainer_email='stefan@seefeld.name',
      description='Faber is a construction tool.',
      url='https://stefanseefeld.github.io/faber',
      download_url='https://github.com/stefanseefeld/faber/releases',
      license='BSL',
      classifiers=['Environment :: Console',
                   'Intended Audience :: Developers',
                   'License :: OSI Approved :: Boost Software License 1.0 (BSL-1.0)',
                   'Operating System :: OS Independent',
                   'Topic :: Software Development :: Build Tools',
                   'Topic :: Software Development :: Testing',
                   'Programming Language :: Python'],
      cmdclass=versioneer.get_cmdclass({'build_doc': build_doc}),
      package_dir={'': 'src'},
      packages=find_packages(where='src'),
      entry_points=dict(console_scripts=['faber=faber.cli:cli_main']),
      data_files=data + docs,
      setup_requires=['pytest-runner'],
      tests_require=['pytest'],
      )



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from __future__ import absolute_import
from . import types
from .feature import set
from .artefact import source, intermediate
from .rule import rule as explicit_rule
from .utils import aslist
from collections import defaultdict
import logging

logger = logging.getLogger('rules')


class _candidate(object):

    def __init__(self, recipe, type, target, sources, features, intermediate, scan=None, logfile=None):
        self.recipe = recipe
        self.type = type
        self.target = target
        self.sources = sources
        self.features = features
        self.intermediate = intermediate
        self.scan = scan  # this might be a recipe for include-scanning
        self.logfile = logfile

    def instantiate(self, module):
        """Apply the recipe to the given artefact and source object."""

        sources = [s.instantiate(module) for s in self.sources]
        t = explicit_rule(self.recipe, self.target, sources=sources,
                          features=self.features,
                          attrs=intermediate if self.intermediate else 0, module=module, logfile=self.logfile)
        if self.scan:
            from .artefacts.include_scan import scan
            for s in sources:
                scan(s, t, self.scan, features=self.features, module=module)
        return t

    def __repr__(self):
        return '{} <- {}'.format(self.target, self.sources)


class _noop(_candidate):

    def __init__(self, type, target):
        _candidate.__init__(self, None, type, target, [], None, False)

    def instantiate(self, module):
        return self.target

    def __repr__(self):
        return repr(self.target)


class _implicit_rule(object):
    """An implicit rule instantiates transformations by binding artefact and
    sources."""

    def __init__(self, recipe, target, source, features):
        """Define an implicit rule to map source types to target types.
        Arguments:
          * recipe: the action to perform
          * target, source: output and input types of this generator
          * features: any features to apply."""

        self.recipe = recipe
        self.target = target
        self.source = source if isinstance(source, tuple) else (source,)
        self.features = features

    def bind(self, t, source, features, intermediate, logfile):
        """Instantiate the rule by binding target and source to the recipe.

        Arguments:
        * t: typed_name or artefact
        * source: list of candidates
        * features: a feature set
        * intermediate: boolean
        * logfile:
        """
        # unwrap artefact and source
        target = t.name if type(t) is types.typed_name else t
        src = [s.name if type(s) is types.typed_name else s for s in source]
        logger.info('bind {}: {} <- {}'.format(self.recipe.qname, target, src))

        fs = self.features.copy()
        fs |= features
        # hack: if this is a C/C++ source compilation, inject a header-scan
        #       to track the additional header dependencies
        scan = None
        if source[0].type in (types.c, types.cxx) and t.type is types.obj:
            tool = self.recipe.tool
            scan = tool.makedep if tool else None
        return _candidate(self.recipe, t.type, target, src, fs, intermediate, scan=scan, logfile=logfile)

    def __repr__(self):
        return '<{} {} <- {}>'.format(self.recipe.qname,
                                      self.target.name,
                                      ','.join([s.name for s in self.source]))


_repository = defaultdict(list)


def init():
    _repository.clear()


class NoRuleError(NotImplementedError): pass


def connect(target, source, features, intermediate=False, logfile=None):
    """Create a chain of implicit rules that maps the given sources
    to the requested target."""

    # Find all rules able to produce the requested target...
    rules = _repository[target.type]
    # ...then iterate over those that match the given features
    for r in iter(r for r in rules if r.features.matches(features)):
        try:
            f = features.copy()
            f |= r.features  # propagate feature requirements up

            # now transform the sources to the expected input type(s).
            def recurse(e, s, f):
                if s.type in e:
                    return _noop(s.type, s.name if type(s) is types.typed_name else s)
                else:
                    # try to generate the first expected type from the source
                    etype = e[0]
                    target = etype.typed_name(etype.synthesize_name(s.name))
                    return connect(target, [s], f, intermediate=True, logfile=logfile)
            source = [recurse(r.source, s, f) for s in source]
            return r.bind(target, source, f, intermediate, logfile=logfile)
        except NoRuleError:
            pass
    # If we didn't find a match, give up.
    raise NoRuleError('No implicit rule to generate {} {} from {} matching {}'
                      .format(target.type.name,
                              target.name,
                              [s.name for s in source],
                              features.essentials()))


def implicit_rule(recipe, target, source):
    """Define an implicit rule to build target type from source type using recipe."""

    fs = recipe._tool.features if recipe._tool else set()
    r = _implicit_rule(recipe, target, source, fs)
    _repository[target].append(r)
    logger.info('irule {}: {} <- {} ({})'
                .format(recipe.qname, target, source, fs))


def rule(target, sources, features=set(), intermediate=False, module=None):
    """Construct a rule using a chain of implicit rules to build target from source."""

    # convert target and source to the proper types...
    sources = aslist(sources)
    sources = [source.instantiate(s, module) for s in sources]
    features = set.instantiate(features)
    target = types.type.typed_name(target) if type(target) is str else target

    logger.info('assembly rule: {} <- {}'.format(target, sources))
    # now look at the source types to see what tools we may need.
    for t in (types.c, types.cxx):
        if any([s.type is t for s in sources]):
            from .tools.compiler import compiler
            compiler.check_instance_for_type(t)

    # now build the chain
    chain = connect(target, sources, features, intermediate, logfile=target.logfile)
    return chain.instantiate(module)



#
# Copyright (c) 2018 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from .artefact import flag
import os
import graphviz


def walk(a, d=None):
    """Traverse the prerequisites of `a`."""

    # This of course assumes the graph not to have any cycles.
    yield a, d
    for p in a.prerequisites:
        yield from walk(p, a)


def visualize(*args, filename='dependencies', format=None):
    """Visualize the dependency graph for artefacts in `args`."""

    g = graphviz.Digraph(graph_attr={'rankdir': 'TB'})

    nodes = set()
    edges = set()
    for arg in args:
        for t, s in walk(arg):
            if t not in nodes:
                fillcolor='bisque1:bisque3' if t.flags & flag.NOTFILE else 'darkolivegreen1:darkolivegreen3'
                g.node(str(t), label=t.name, shape='rectangle', style='filled', fillcolor=fillcolor)
                nodes.add(t)
            # Avoid circular references
            if s and s != t and (s, t) not in edges:
                g.edge(str(s), str(t))
                edges.add((s, t))

    fmts = ['.png', '.pdf', '.dot', '.svg', '.jpeg', '.jpg']
    if format is None and any(filename.lower().endswith(fmt) for fmt in fmts):
        filename, format = os.path.splitext(filename)
        format = format[1:].lower()

    if format is None:
        format = 'png'

    data = g.pipe(format=format)
    if not data:
        raise RuntimeError('Graphviz failed to properly produce an image.')

    full_filename = '.'.join([filename, format])
    with open(full_filename, 'wb') as f:
        f.write(data)
    print('output written to {}'.format(full_filename))



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.action import action
from faber.artefact import artefact, notfile, always, nocare, intermediate
from faber.rule import rule, depend
from faber.tools import fileutils
from faber import scheduler
from test.common import pyecho
from os.path import exists
import pytest
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


@pytest.mark.usefixtures('module')
def test_action():
    """Check that an action's status is reported upon completion."""
    a = artefact('a', attrs=notfile|always)
    b = artefact('b', attrs=notfile)
    c = artefact('c', attrs=notfile)
    a = rule(pyecho, a)
    b = rule(pyecho, b, a)
    c = rule(pyecho, c, b)
    with patch('faber.action.action.__status__') as recipe:
        assert b.update()
        (_, _, _, _, output, _), kwds = recipe.call_args_list[-1]
        assert output.strip() == 'b <- a'
        assert c.update()
        (_, _, _, _, output, _), kwds = recipe.call_args_list[-1]
        assert output.strip() == 'c <- b'


@pytest.mark.usefixtures('module')
def test_noop():
    """Check that an artefact won't be updated if a dependent artefact is up to date."""
    a = artefact('a', attrs=notfile)
    b = artefact('b', attrs=notfile)
    b = rule(pyecho, b, a)
    with patch('faber.action.action.__status__') as recipe:
        assert b.update()
        assert not recipe.called


@pytest.mark.usefixtures('module')
def test_fail():
    """Check that an artefact won't be updated if a dependent artefact's recipe failed."""
    a = artefact('a', attrs=notfile|always)
    b = artefact('b', attrs=notfile)
    fail = action('failing', 'fail')
    a = rule(fail, a)
    b = rule(pyecho, b, a)
    with patch('faber.action.action.__status__'):
        assert not b.update()


@pytest.mark.usefixtures('module')
def test_nocare():
    """Check that an artefact will be updated if a dependent artefact's recipe failed
    but was marked as nocare."""
    a = artefact('a', attrs=notfile|always|nocare)
    b = artefact('b', attrs=notfile)
    fail = action('failing', 'fail')
    a = rule(fail, a)
    b = rule(pyecho, b, a)
    with patch('faber.action.action.__status__') as recipe:
        assert b.update()
        (_, _, _, _, output, _), kwds = recipe.call_args_list[-1]
        assert output.strip() == 'b <- a'


@pytest.mark.usefixtures('module')
def test_dynamic_dependencies():
    """Test whether it's possible to add dependencies while the scheduler is already running."""

    c = artefact('c', attrs=notfile)

    def inject_deps(self, *args, **kwds):
        d=rule(pyecho, 'd', attrs=notfile|always)
        depend(c, d)
        print('ddeps.b')

    a = rule(pyecho, 'a', attrs=notfile|always)
    b = rule(action('dg', inject_deps), 'b', a, attrs=notfile)
    c = rule(pyecho, c, b)
    with patch('faber.action.action.__status__') as recipe:
        assert b.update()
        (_, _, _, _, output, _), kwds = recipe.call_args_list[-1]
        assert output.strip() == 'ddeps.b'
        assert c.update()
        # verify that d is actually updated before c
        output = [i[0][4].strip() for i in recipe.call_args_list]
        assert 'd <-' in output


@pytest.mark.usefixtures('module')
def test_dynamic_recipe():
    """Test whether it's possible to add a recipe while the scheduler is already running."""

    c = artefact('c', attrs=notfile)

    def generate(*args, **kwds):
        """generate the graph for c."""
        a = rule(pyecho, 'a', attrs=notfile|always)
        a1 = rule(pyecho, 'a1', a, attrs=notfile|always)
        a2 = rule(pyecho, 'a2', a1, attrs=notfile|always)
        rule(pyecho, c, a2, )

    b = rule(generate, 'b', attrs=notfile|always)
    depend(c, b)
    with patch('faber.action.action.__status__') as recipe:
        assert c.update()
        output = [i[0][4].strip() for i in recipe.call_args_list]
        assert output[-1] == 'c <- a2'


@pytest.mark.usefixtures('module')
def test_multi():
    """TBD"""

    # workflow:
    # b1 needs to be updated, but doing that also updates b2
    # this means c2 (dependent on b2) will also be updated.
    a = artefact('a', attrs=notfile)
    b1 = artefact('b1', attrs=notfile)
    b2 = artefact('b2', attrs=notfile|always)
    c1 = artefact('c1', attrs=notfile)
    c2 = artefact('c2', attrs=notfile)
    d = artefact('d', attrs=notfile)
    rule(pyecho, a)
    rule(pyecho, [b1, b2], a)
    rule(pyecho, c1, b1)
    rule(pyecho, c2, b2)
    rule(pyecho, d, [c1, c2])
    with patch('faber.action.action.__status__') as recipe:
        assert d.update()
        output = [i[0][4].strip() for i in recipe.call_args_list]
        assert output[0] == 'b1 b2 <- a'
        assert 'c2 <- b2' in output
        # sets aren't ordered, so we check for both possibilities
        assert output[-1] in ('d <- c1 c2', 'd <- c2 c1')


@pytest.mark.usefixtures('module')
def test_intermediate():
    """Test that intermediate files are properly updated."""

    a = rule(fileutils.touch, 'a', attrs=intermediate)
    b = rule(fileutils.touch, 'b', a, attrs=intermediate)
    c = rule(pyecho, 'c', b, attrs=notfile|always)
    assert c.update()
    assert exists(a._filename)
    assert exists(b._filename)


@pytest.mark.usefixtures('module')
def test_late():
    """Test that a "late" dependency raises an error."""

    a = artefact('a', attrs=notfile|always)
    assert a.update()
    with pytest.raises(scheduler.DependencyError):
        b = artefact('b', attrs=notfile)
        depend(a, b)


@pytest.mark.usefixtures('module')
def test_cycle():
    """Test that the scheduler detects dependency cycles."""

    echo = action('echo', 'echo $(<) $(>)')
    a = artefact('a', attrs=notfile|always)
    b = rule(echo, 'b', a, attrs=notfile)
    with pytest.raises(scheduler.DependencyError):
        a = rule(echo, a, b)


@pytest.mark.usefixtures('module')
def test_late_cycle():
    """Test that the scheduler detects dependency cycles
    created in recipes."""

    a = artefact('a', attrs=notfile|always)
    b = artefact('b', attrs=notfile|always)
    c = artefact('c', attrs=notfile|always)

    def generator(targets, sources):
        depend(b, c)  # create cycle !
    b = rule(generator, b, a)
    echo = action('echo', 'echo $(<) $(>)')
    c = rule(echo, c, b)
    with pytest.raises(scheduler.DependencyError):
        assert c.update()



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

pass_ = 0x1
fail = 0x2
xpass = 0x3
xfail = 0x4

from .test import test  # noqa F401
from .suite import suite  # noqa F401
from .report import report  # noqa F401



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from contextlib import contextmanager
import string
import sys


def add_metaclass(metaclass):

    # This is a minimalist version to satisfy faber's needs only
    def wrapper(cls):
        vars = cls.__dict__.copy()
        vars.pop('__dict__', None)
        vars.pop('__weakref__', None)
        return metaclass(cls.__name__, cls.__bases__, vars)
    return wrapper


@contextmanager
def capture_output():
    if sys.version_info < (3, 0):
        from io import StringIO

        class TextIO(StringIO):
            def write(self, data):
                if not isinstance(data, unicode):  # noqa F821
                    data = unicode(data, getattr(self, '_encoding', 'utf-8'), 'replace')  # noqa F821
                StringIO.write(self, data)
    else:
        from io import StringIO as TextIO
    out = TextIO()
    err = TextIO()
    stdout = sys.stdout
    sys.stdout = out
    stderr = sys.stderr
    sys.stderr = err
    try:
        yield out, err
    finally:
        sys.stdout = stdout
        sys.stderr = stderr


def aslist(o):
    return o if isinstance(o, list) else [o]


class path_formatter(string.Formatter):

    def format_field(self, value, spec):
        """Remove path segments corresponding to undefined features."""
        if value is None:
            return ''
        else:
            return super(path_formatter, self).format_field(value, spec)
    def get_value(self, key, args, kwargs):
        """Remove path segments corresponding to non-existent features."""
        try:
            if isinstance(key, int):
                return args[key]
            else:
                return kwargs[key]
        except KeyError:
            return ''

    def get_field(self, field_name, args, kwargs):
        if sys.version_info < (3, 0):
            first, rest = field_name._formatter_field_name_split()
        else:
            first, rest = string._string.formatter_field_name_split(field_name)
        try:
            # if `first` doesn't exist, this throws
            obj = string.Formatter.get_value(self, first, args, kwargs)
            # loop through the rest of the field_name, doing
            #  getattr or getitem as needed
            for is_attr, i in rest:
                if is_attr:
                    obj = getattr(obj, i)
                else:
                    obj = obj[i]

            return obj, first
        except Exception:
            return '', first



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from ..artefact import artefact, source, notfile, always, nopropagate
from ..rule import rule, depend
from ..action import action
from ..utils import aslist
from .. import assembly
from .. import output
import logging
import sys

action_logger = logging.getLogger('actions')
command_logger = logging.getLogger('commands')


class assemble(action):

    def call(target, source, **kwds):
        self = composite._targets[target[0]]
        # create the real dependency graph...
        self._assemble()

    # run even in 'noexec' mode.
    # (see ..scheduler._pyaction)
    call.__noexec__ = True
    command = staticmethod(call)

    def __status__(self, targets, status, command, time, stdout, stderr):
        # suppress output unless we are in debug mode
        if action_logger.isEnabledFor(logging.DEBUG):
            if isinstance(targets, list):
                targets = ' '.join([t.qname for t in targets])
            else:
                targets = targets.qname
            action_logger.debug(output.coloured('{} {}'.format(self.qname, targets), attrs=['bold']))
        command_logger.debug(command)
        if not status:
            print(stderr, file=sys.stderr)


class composite(artefact):
    """A composite artefact is built in stages using one or more intermediates."""

    _targets = {}  # map assembler artefact to composite

    def __init__(self, name, sources, *args, **kwds):
        dependencies = kwds.pop('dependencies', [])
        artefact.__init__(self, name, *args, **kwds)
        self.sources = [source.instantiate(a, self.module) for a in aslist(sources)]
        self.dependencies = aslist(dependencies)
        self.features |= artefact.combine_use(self.sources)
        if self.features.dependencies():
            # postpone assembly until after dependencies are updated
            self._submit()
        else:
            # without any delayed values, assemble immediately
            self._assemble()

    def reset(self):
        # do not reset this artefact as we can't undo the assembly
        pass

    def __call__(self, features):
        c = artefact.__call__(self, features)
        if c.features.dependencies():
            c._submit()
        else:
            c._assemble()
        return c

    def _submit(self):
        a = rule(assemble(), 'a:' + self.name, attrs=notfile|always|nopropagate,
                 dependencies=self.sources + self.dependencies + self.features.dependencies())
        composite._targets[a] = self
        depend(self, a)

    def _assemble(self):
        from ..tools import compiler  # noqa F401
        if self.status is None:
            self.features.eval(update=False)
            assembly.rule(self, self.sources, self.features, module=self.module)



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

import sqlite3
import hashlib
import os
import os.path


class filecache(object):
    """Record all file artefacts to facilitate their cleanup."""

    def __init__(self, builddir, params):
        self.root = builddir
        fdir = os.path.join(builddir, '.faber')
        if not os.path.exists(fdir):
            os.makedirs(fdir)
        self.filename = os.path.join(fdir, 'filecache')
        variant = str(params).encode('utf-8')
        self.variant = hashlib.md5(variant).hexdigest()
        self.conn = sqlite3.connect(self.filename)
        # Create table if it doesn't exist yet.
        if not next(self.conn.execute('SELECT name FROM sqlite_master '
                                      'WHERE type="table" AND name="files"'), None):
            self.conn.execute('CREATE TABLE files (variant TEXT, filename TEXT)')

    def __del__(self):
        self.conn.commit()

    def append(self, filename):
        with self.conn:
            self.conn.execute('INSERT INTO files VALUES(?,?)',
                              (self.variant, filename))

    def extend(self, filenames):
        with self.conn:
            self.conn.executemany('INSERT INTO files VALUES(?,?)',
                                  [(self.variant, f) for f in filenames])

    def clear(self):
        with self.conn:
            self.conn.execute('DELETE FROM files')

    def __iter__(self):
        with self.conn:
            for i in self.conn.execute('SELECT filename FROM files WHERE variant=?',
                                       (self.variant,)):
                yield i[0]



#
# Copyright (c) 2018 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from __future__ import absolute_import
import logging
from logging import getLogger

topics = dict(summary=1,    # build summary
              actions=2,    # action names
              commands=4,   # executed commands
              output=8,     # command output
              process=0,    # scheduler backend
              scheduler=0,  # interaction with scheduler
              tools=0,      # instantiation of tools
              features=0,   # creation and manipulation of features
              rules=0,      # declaration of rules (including implicit ones)
              config=0,     # any config-related activity
              test=0)       # any testing-related activity


class ProfileFormatter(logging.Formatter):
    """If the given record contains a `time` field and it is a non-negative
    value, add its value to the log output."""

    def format(self, record):
        s = super(ProfileFormatter, self).format(record)
        time = record.time if hasattr(record, 'time') else -1.
        if time >= 0.:
            s += '\t (time={} s)'.format(time)
        return s


def setup(log=None, loglevel=None, debug=False, profile=False):
    """If `log` is a list of loggers to enable, or None to indicate default settings."""

    # if debugging is enabled, print everything...
    if debug:
        logging.basicConfig(format='%(name)s: %(message)s', level=logging.DEBUG)
    else:
        logging.basicConfig(format='%(message)s')
        if log:  # if loggers are set by name...
            for l in log:
                getLogger(l).setLevel(level=logging.INFO)
        elif loglevel is not None:  # if loggers are enabled by level...
            for n, l in topics.items():
                if loglevel & l:
                    getLogger(n).setLevel(level=logging.INFO)
        elif log is None:  # undefined, use defaults
            # by default, enable summary, actions, and output loggers only
            getLogger('summary').setLevel(level=logging.INFO)
            getLogger('actions').setLevel(level=logging.INFO)
            getLogger('output').setLevel(level=logging.INFO)
    if profile:
        h = logging.StreamHandler()
        h.setFormatter(ProfileFormatter())
        c = getLogger('commands')
        c.handlers[:] = []  # hack !
        c.addHandler(h)
        c.propagate = False
        # `-p` implies `--log=commands`
        c.setLevel(level=logging.INFO)



#
# Copyright (c) 2016 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.feature import *
import pytest


def test_link():

    link = feature('link', ('shared', 'static'))

    l1 = link('shared')
    l2 = link('static')
    with pytest.raises(ValueError):
        l3 = link('invalid')  # noqa F841
    assert l1.matches(l1)
    assert not l1.matches(l2)
    assert not l2.matches(l1)


def test_include():

    include = feature('include', attributes=multi|incidental)
    i1 = include('a', 'b')
    i2 = include('c')
    i1 += i2
    assert i1 == ('a', 'b', 'c')
    assert i1.matches(i2)
    assert i2.matches(i1)


def test_composite():

    tool = feature('tool', feature('name', sub=True), feature('version', sub=True))
    t1 = tool(name='g++')
    t2 = tool(name='g++', version='6.3')
    assert t1 != t2
    assert t1.name == t2.name
    assert t1.version != t2.version
    assert t1.matches(t2)
    assert t2.matches(t1)
    t3 = tool(name='clang++')
    assert t1 != t3
    assert t1.name != t3.name
    assert not t1.matches(t3)

    some = feature('some', feature('include', sub=True, attributes=multi))
    s1 = some(include=('a', 'b'))
    include = feature('include', attributes=multi)
    i1 = include('a', 'b')
    assert s1.include != i1

    s2 = some(include=('a', 'c'))
    s1 += s2
    assert s1.include == ('a', 'b', 'a', 'c')


def test_parsing():

    tool = feature('tool', feature('name', sub=True), feature('version', sub=True))
    include = feature('include', attributes=multi)
    define = feature('define', attributes=multi)
    gxx = tool(name='g++', version='6.3')
    clangxx = tool(name='clang++', version='3.9')
    variant = feature('variant', ('release', 'debug'))  # noqa F841

    o1 = {'tool.name': 'g++', 'variant': 'release'}
    o2 = {'tool.name': 'clang++', 'tool.version': '3.9', 'variant': 'release'}
    fs = lazy_set(o1)
    t1 = fs.tool
    v1 = fs.variant
    fs = lazy_set(o2)
    t2 = fs.tool
    v2 = fs.variant
    assert gxx.matches(t1)
    assert v1 == 'release'
    assert not clangxx.matches(t1)
    assert not gxx.matches(t2)
    assert clangxx.matches(t2)
    assert v2 == 'release'

    o3 = {'include': 'AB'}
    fs = lazy_set(o3)
    assert fs.include == ('AB',)
    o4 = {'include': ['A', 'B']}
    fs = lazy_set(o4)
    assert fs.include == ('A', 'B')

    o5 = {'define': 'C', 'new_tool.name': 'g++', 'new_tool.version': '6.3'}
    lazy = lazy_set(o5, define('A', 'B'), include('I'))
    assert 'define' in lazy
    assert lazy.define == ('C', 'A', 'B')
    assert 'new_tool' not in lazy
    tool = feature('new_tool', feature('name', sub=True), feature('version', sub=True))
    assert 'new_tool' in lazy
    assert lazy.new_tool.name == 'g++'

    o6 = {'tool': 'g++-7'}
    fs = lazy_set(o6)
    assert fs.tool.name == 'g++' and fs.tool.version == '7'


def test_parsing_conflict():

    tool = feature('tool', feature('name', sub=True), feature('version', sub=True))  # noqa F841

    o1 = {'tool': 'g++-2', 'tool.version': '1'}
    fs = lazy_set(o1)
    with pytest.raises(ValueError):
        # trying to instantiate the values should fail
        fs.tool.name == 'g++' and fs.tool.version == 2


def test_serialize():

    tool = feature('tool', feature('name', sub=True), feature('version', sub=True))  # noqa F841
    variant = feature('variant', ('release', 'debug'))  # noqa F841
    o = {'tool.name': 'clang++', 'tool.version': '3.9', 'variant': 'release'}
    fs = lazy_set(o)
    d = fs._serialize()
    assert d == o


def test_mix():

    link1 = feature('link1', ('shared', 'static'))
    link2 = feature('link2', ('shared', 'static'))
    l1 = link1('shared')
    l2 = link2('shared')
    l3 = link1('shared')
    assert l1 != l2
    assert not l1 == l2
    assert l1 == l3
    assert not l1 != l3
    with pytest.raises(ValueError):
        l1 += l2


def test_feature_set():

    tool = feature('tool', feature('name', sub=True), feature('version', sub=True))
    target = feature('target', feature('arch', sub=True))
    link = feature('link', ('shared', 'static'))
    include = feature('include', attributes=multi|incidental)
    linkpath = feature('linkpath', attributes=multi|incidental)

    fs = set(link('shared'), include('a', 'b'))
    fs += set(include('c'), linkpath('d'))
    fs += tool(name='g++', version='6.3')
    fs += target(arch='A')

    assert 'link' in fs
    assert 'define' not in fs

    assert fs.tool.name == 'g++'
    assert fs.link == 'shared'
    assert fs.include == ('a', 'b', 'c')
    assert fs.linkpath == ('d',)

    fs.update(link('static'))
    assert fs.link == 'static'

    assert len(fs) == 5

    fs2 = fs.copy()
    assert fs2.matches(fs)
    del fs2.link
    assert fs2.matches(fs)
    assert fs.matches(fs2)
    fs2.include += include('d')
    assert fs2.matches(fs)
    assert fs.matches(fs2)

    del fs2.target
    assert fs2.matches(fs)
    assert fs.matches(fs2)


def test_fs_copy():

    include = feature('include', attributes=multi)
    link = feature('link', ['shared', 'static'])
    define = feature('define', attributes=multi)
    fs = set(include('a', 'b'), link('shared'))
    fs += define('SHARED', condition=set.link == 'shared')
    assert fs.include == ('a', 'b')
    fs2 = fs.copy()
    fs2 += include('c')
    assert fs.include == ('a', 'b')
    assert fs2.include == ('a', 'b', 'c')

    i = include('a')
    ls = lazy_set({})
    ls |= i
    ls2 = ls.copy()
    ls |= include('c')
    assert ls.include == ('a', 'c')
    assert ls2.include == 'a'

    i = include('a')
    ls = lazy_set({})
    ls |= i
    ls2 = ls.copy()
    print(id(ls.include._value), id(ls2.include._value))
    ls |= include('c')
    print(id(ls.include._value), id(ls2.include._value))
    assert ls.include == ('a', 'c')
    assert ls2.include == 'a'


def test_mapping():

    define = feature('define', attributes=multi)
    include = feature('include', attributes=multi)
    link = feature('link', ('shared', 'static'))
    linkpath = feature('linkpath', attributes=multi)
    libs = feature('libs', attributes=multi)

    fs = set(link('shared'), include('a', 'b'), linkpath('c'), libs('foo', 'bar'))

    cppflags = map(include, translate, prefix='-I') + map(define, translate, prefix='-D')
    cxxflags = map(link, select_if, 'shared', '-fPIC')
    linkpath = map(linkpath, translate, prefix='-L')
    libs = map(libs, translate, prefix='-l', suffix='.lib')

    assert cppflags(fs) == '-Ia -Ib'
    assert cxxflags(fs) == '-fPIC'
    assert linkpath(fs) == '-Lc'
    assert libs(fs) == '-lfoo.lib -lbar.lib'


def test_delayed():

    tool = feature('tool', feature('name', sub=True), feature('version', sub=True))
    define = feature('define', attributes=multi)
    include = feature('include', attributes=multi)
    link = feature('link', ['shared', 'static'])
    fs = set(define('MACRO'), include('a', 'b'), link('shared'),
             tool(name='g++', version='6.3'))
    c1 = set.link == 'shared'
    c2 = set.link == 'static'
    c3 = set.tool.name == 'g++'
    c4 = set.tool.name == 'clang++'
    # doesn't work due to language restrictions
    # c5 = 'MACRO' in set.define
    c5 = set.define.contains('MACRO')
    c6 = c5.not_()
    c7 = set.nonexistent.contains('nothing')
    with pytest.raises(ValueError):
        r1 = c3 or c4
    with pytest.raises(ValueError):
        r2 = c3 and c4
    r1 = c3 | c4
    r2 = c3 & c4
    assert c1(fs) is True
    assert c2(fs) is False
    assert c3(fs) is True
    assert c5(fs) is True
    assert c6(fs) is False
    assert c7(fs) is False
    assert r1(fs) is True
    assert r2(fs) is False

    cc1 = c1 | c2
    cc2 = c1 & c2
    assert cc1(fs) is True
    assert cc2(fs) is False

    c3 = set.foo == 'bar'
    assert c3(fs) is False

    fs += define('SHARED', condition=set.link == 'shared')
    fs += define('GXX', condition=set.tool.name == 'g++')
    fs += define('CLANGXX', condition=set.tool.name == 'clang++')
    fs += define('CXX', condition=((set.tool.name == 'clang++') |
                                   (set.tool.name == 'g++')))
    fs += include('/some/path', condition=set.tool.name == 'g++')
    assert fs.define == ('MACRO',)
    fs.eval()
    assert fs.define == ('MACRO', 'SHARED', 'GXX', 'CXX')
    assert fs.include == ('a', 'b', '/some/path',)


def test_condition():

    tool = feature('tool', feature('name', sub=True), feature('version', sub=True))
    target = feature('target', feature('os', sub=True), feature('arch', sub=True))
    define = feature('define', attributes=multi)
    fs = set(tool(name='g++', version='6.3'), target(os='gnu-linux', arch='x86_64'))
    fs += define('LINUX', condition=set.target.os.matches('.*-linux'))
    fs.eval()
    assert fs.define == ('LINUX',)



#
# Copyright (c) 2018 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.scheduler.artefact import artefact
from faber.scheduler.recipe import recipe
import pytest
import tempfile
import shutil


@pytest.fixture()
def scheduler():
    artefact.init()
    recipe.init()
    yield
    recipe.finish()
    artefact.finish()


@pytest.fixture()
def tempdir():
    dirpath = tempfile.mkdtemp()
    try: yield dirpath
    finally: shutil.rmtree(dirpath)



#
# Copyright (c) 2018 Stefan Seefeld
# All rights reserved.
#
# This file is part of Faber. It is made available under the
# Boost Software License, Version 1.0.
# (Consult LICENSE or http://www.boost.org/LICENSE_1_0.txt)

from faber.scheduler.artefact import *
from . import make_artefact, touch
import pytest
from os.path import exists, join


@pytest.mark.asyncio
@pytest.mark.usefixtures('scheduler')
async def test_init():
    a = make_artefact('a')
    assert a.fate == fate.INIT


@pytest.mark.asyncio
@pytest.mark.usefixtures('scheduler')
async def test_binding():
    """Test an individual artefact's binding state before
    and after calling `a.bind()`."""

    a = make_artefact(name=__file__)
    # initially an artefact is unbound...
    assert a.binding == binding.UNBOUND, 'a is already bound'
    await a.bind()
    # ...binding it will determine whether it exists
    # or is missing.
    assert a.binding == binding.EXISTS, 'a does not exist'
    # generate a filename we know doesn't exist
    b = make_artefact(name=__file__.replace('.py', '.so'))
    await b.bind()
    assert b.binding == binding.MISSING, 'b is not missing'
    c = make_artefact(name='c', attrs=flag.NOTFILE)
    await c.bind()
    # non-file artefacts are never bound
    assert c.binding == binding.UNBOUND, 'c is not bound yet'


@pytest.mark.asyncio
@pytest.mark.usefixtures('scheduler')
async def test_process(tempdir):
    # nothing to do
    a = make_artefact(join(tempdir, 'a'))
    await a.process()
    assert a.progress == progress.DONE, 'a is still incomplete'
    # nothing to do
    b = make_artefact(join(tempdir, 'b'), touch=True)
    await b.process()
    assert b.progress == progress.DONE, 'b is still incomplete'
    assert exists(b.boundname), 'b does not exist'


@pytest.mark.asyncio
@pytest.mark.usefixtures('scheduler')
async def test_simple_dep(tempdir):
    """Test that an artefact is correctly updated including its prerequisites"""
    a = make_artefact(join(tempdir, 'a'), touch=True)
    b = make_artefact(join(tempdir, 'b'), touch=True, prerequisites=[a])
    c = make_artefact(join(tempdir, 'c'), touch=True, prerequisites=[b])
    await c.process()
    assert c.progress == progress.DONE, 'c is still incomplete'
    assert exists(c.boundname) and c.recipe.status, 'c was not updated'


@pytest.mark.asyncio
@pytest.mark.usefixtures('scheduler')
async def test_simple_dep2(tempdir):
    """Test that an artefact is only updated if it is out-of-date"""
    a = make_artefact(join(tempdir, 'a'), touch=True)
    b = make_artefact(join(tempdir, 'b'), touch=True, prerequisites=[a])
    c = make_artefact(join(tempdir, 'c'), touch=True, prerequisites=[b])
    touch(a.name)
    await asyncio.sleep(1)
    touch(b.name)
    await c.process()
    assert c.progress == progress.DONE, 'c is still in progress'
    assert exists(c.boundname) and c.recipe.status
    assert a.recipe.status is None, 'a was wrongly updated'
    assert b.recipe.status is None, 'b was wrongly updated'


@pytest.mark.asyncio
@pytest.mark.usefixtures('scheduler')
async def test_notfile_dep(tempdir):
    a = make_artefact(join(tempdir, 'a'), touch=True)
    b = make_artefact('b', attrs=flag.NOTFILE, prerequisites=[a])
    await b.process()
    assert b.progress == progress.DONE, 'b is still in progress'


@pytest.mark.asyncio
@pytest.mark.usefixtures('scheduler')
async def test_temp_dep(tempdir):
    a = make_artefact(join(tempdir, 'a'), touch=True)
    b = make_artefact(join(tempdir, 'b'), attrs=flag.TEMP, touch=True, prerequisites=[a])
    c = make_artefact(join(tempdir, 'c'), touch=True, prerequisites=[b])
    await c.process()
    assert exists(a.boundname) and a.recipe.status, 'a was not updated'
    assert exists(b.boundname) and b.recipe.status, 'b was not updated'
    assert exists(c.boundname) and c.recipe.status, 'c was not updated'


@pytest.mark.asyncio
@pytest.mark.usefixtures('scheduler')
async def test_temp_dep2(tempdir):
    a = make_artefact(join(tempdir, 'a'), touch=True)
    b = make_artefact(join(tempdir, 'b'), attrs=flag.TEMP, touch=True, prerequisites=[a])
    c = make_artefact(join(tempdir, 'c'), touch=True, prerequisites=[b])
    # update a before c, so c is up-to-date
    touch(a.name)
    await asyncio.sleep(1)
    touch(c.name)
    await c.process()
    assert b.recipe.status is None, 'b was wrongly updated'
    assert c.recipe.status is None, 'c was wrongly updated'


@pytest.mark.asyncio
@pytest.mark.usefixtures('scheduler')
async def test_temp_dep3(tempdir):
    a = make_artefact(join(tempdir, 'a'), touch=True)
    b = make_artefact(join(tempdir, 'b'), attrs=flag.TEMP, touch=True, prerequisites=[a])
    c = make_artefact(join(tempdir, 'c'), touch=True, prerequisites=[b])
    # update c before a, so c needs to be updated
    touch(c.name)
    await asyncio.sleep(1)
    touch(a.name)
    await c.process()
    assert b.recipe.status, 'b was not updated'
    assert c.recipe.status, 'c was not updated'
